/* Copyright (C) 1998,1999,2000 Nikos Mavroyanopoulos
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU Library General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/* $Id: mcrypt.c,v 1.16 2000/03/01 13:12:32 nikos Exp $ */

/* Changed by Steve Underwood 1999/12/10 to allow an arbitrary number of 
 * streams of encryption. Currently the resulting code is probably not
 * thread safe, but as far as I could tell the previous code wasn't
 * either. This version has brute force locking in a lot of places, but
 * it has not been tested in a multi-threaded manner.
 * The key locking issue is that the table of encryption streams could
 * be moved when it is extended. Any address pre-calculated, or in
 * calculation at the time of the reallocation would be screwed.
 * This won't happen often, but requires lots of locks - PITA! 
 */

/* Changed again at 1999/12/15 to correct the thread safeness. Now it
 * seems to be thread safe. Brute force locking was removed and
 * locks per thread were introduced. 
 *						--nikos
 */

#ifndef LIBDEFS_H
#define LIBDEFS_H
#include <libdefs.h>
#endif
#include <bzero.h>
#include <mcrypt.h>
#include <xmemory.h>
#include <mcrypt_internal.h>

#ifdef HAVE_LIBPTHREAD
pthread_mutex_t counter_stream_lock = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t lt_mutex_lock = PTHREAD_MUTEX_INITIALIZER;
#endif

CRYPT_STREAM *streams = NULL;
int max_streams_to_date = 0;
#if 0
int preloaded_symbols = 0;
#endif

int internal_init_mcrypt(int td, void *key, int lenofkey, void *IV)
{
	int *sizes=NULL;
	int num_of_sizes, i, ok = 0;

	if (max_streams_to_date == 0)
		return MCRYPT_UNKNOWN_ERROR;	/* error */

	if (lenofkey == 0) {
		return MCRYPT_KEY_LEN_ERROR;	/* error */
	}

	sizes = mcrypt_get_supported_key_sizes(td, &num_of_sizes);
	if (sizes!=NULL) {
		for (i = 0; i < num_of_sizes; i++) {
			if (lenofkey == sizes[i]) {
				ok = 1;
				break;
			}
		}
	} else { /* sizes==NULL */
		if (num_of_sizes==0) ok=1;
	}
	if (ok == 0) {
		free(sizes);
		return MCRYPT_KEY_LEN_ERROR;
	}
	free(sizes);

	LOCK;
	streams[td].keyword_given = mxcalloc(1, mcrypt_get_key_size(td));
	memmove(streams[td].keyword_given, key, lenofkey);
	streams[td].akey = mxcalloc(1, mcrypt_get_size(td));
	streams[td].abuf = mxcalloc(1, mcrypt_mode_get_size(td));

	init_mcrypt(td, streams[td].abuf, key, lenofkey, IV);

	mcrypt_set_key(td,
		       (void *) streams[td].akey,
		       (void *) streams[td].keyword_given,
		       lenofkey, IV, mcrypt_get_iv_size(td));

	if (streams[td].akey==NULL) {
		UNLOCK;
		internal_end_mcrypt(td);
		return MCRYPT_UNKNOWN_ERROR;
	}
	UNLOCK;
	return td;
}

int internal_end_mcrypt(int td)
{
	LOCK;
	mxfree(streams[td].keyword_given, mcrypt_get_key_size(td));
	streams[td].keyword_given = NULL;

	mxfree(streams[td].akey, mcrypt_get_size(td));
	streams[td].akey = NULL;

	end_mcrypt(td, streams[td].abuf);
	mxfree(streams[td].abuf, mcrypt_mode_get_size(td));
	streams[td].abuf = NULL;
	UNLOCK;

	return 0;
}

/* Generic - High level functions */

int  mcrypt_generic_init(const int td, void *key, int lenofkey, void *IV)
{
	return internal_init_mcrypt(td, key, lenofkey, IV);
}

int  mcrypt_generic(int td, void *plaintext, int len)
{
	int x;

	MCRYPT_LOCK;
	x = mcrypt(td, streams[td].abuf, plaintext, len);
	MCRYPT_UNLOCK;
	return x;
}

int  mdecrypt_generic(int td, void *ciphertext, int len)
{
	int x;
	MCRYPT_LOCK;
	x = mdecrypt(td, streams[td].abuf, ciphertext, len);
	MCRYPT_UNLOCK;
	return x;
}

int  mcrypt_generic_end(const int td)
{
	internal_end_mcrypt(td);
	mcrypt_module_close(td);
	return 0;
}


void  mcrypt_perror(int err)
{

	switch (err) {
	case MCRYPT_UNKNOWN_ERROR:
		fprintf(stderr, "Unknown error.\n");
		break;
	case MCRYPT_ALGORITHM_MODE_INCOMPATIBILITY:
		fprintf(stderr,
			"Algorithm incompatible with this mode.\n");
		break;
	case MCRYPT_KEY_LEN_ERROR:
		fprintf(stderr, "Key length cannot be zero.\n");
		break;
	case MCRYPT_MEMORY_ALLOCATION_ERROR:
		fprintf(stderr, "Memory allocation failed.\n");
		break;
	case MCRYPT_UNKNOWN_MODE:
		fprintf(stderr, "Unknown mode.\n");
		break;
	case MCRYPT_UNKNOWN_ALGORITHM:
		fprintf(stderr, "Unknown algorithm.\n");
		break;

	}
	return;
}
